import re

from jiama.constant import EXCHANGE_NAME, REQUEST_QUEUE_NAME, RESPONSE_QUEUE_NAME
from jiama.rmq import Admin


async def list_entity(uri, entity=None):
    entity_queue = {'service': REQUEST_QUEUE_NAME, 'client': RESPONSE_QUEUE_NAME}
    async with Admin().init(uri) as admin:
        qs = await admin.list_queue(columns=['name'], sort='name')
    return {
        e: [
            g.group(1)
            for q in qs
            if (g := re.match(re.sub(r'{.*?}', r'([^.]*)?$', qn), q['name']))
        ]
        for e, qn in entity_queue.items()
        if not entity or entity == e or entity not in entity_queue
    }


async def stats(uri):
    entity_dict = await list_entity(uri)

    async with Admin().init(uri) as admin:
        overview = await admin.overview()
    return {
        'exchange name': EXCHANGE_NAME,
        'service number': len(entity_dict.get('service', [])),
        'client number': len(entity_dict.get('client', [])),
        'broker overview': overview,
    }



import importlib
import inspect
import re

from functools import wraps


# 使用 __init__ 作为工厂方法 不要传入任何参数，
# 使用另外一个方法作为初始化方法接收参数 同时可以实现异步初始化 需提前初始化
def singleton(cls):
    instances = {}

    @wraps(cls)
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance


def load_service(module_path):
    path, _, name = module_path.partition(':')
    module = importlib.import_module(path)
    services = {}
    if name:
        services[name] = getattr(module, name)
    else:
        for name, cls in inspect.getmembers(module, inspect.isclass):
            if inspect.getmembers(cls, is_entrypoint):
                services[name] = cls
    return services


def is_entrypoint(method):
    return getattr(method, 'jiama_rpc_entrypoint', False)


def merge_dict(a, b, ignore=[]):
    if not (isinstance(a, dict) and isinstance(b, dict)):
        if isinstance(a, list) and isinstance(b, list):
            return a + b
        return b

    result = {}
    for k in sorted(set(list(a.keys()) + list(b.keys()))):
        if k in ignore:
            continue
        v1 = a.get(k, None)
        v2 = b.get(k, None)

        if v2:
            result[k] = merge_dict(v1, v2, ignore)
        else:
            result[k] = v1
    return result


def seperate_uri(uri):
    p = r'^(?P<protocol>.+?//)(?P<user>.+?):(?P<password>.+?)@(?P<address>.+)$'
    m = re.match(p, uri)
    return (
        {
            'protocol': m.group('protocol'),
            'address': m.group('address'),
            'user': m.group('user'),
            'password': m.group('password'),
        }
        if m
        else None
    )


def iter_dict(d, i=0, s='    '):
    for k, v in d.items():
        if isinstance(v, dict):
            yield f'{s*i}{k}: '
            yield from iter_dict(v, i + 1, s)
        else:
            yield f'{s*i}{k}: {v}'



import asyncio
import sys

import click
import tomli

from loguru import logger

from jiama.constant import DEFAULT_CONFIG
from jiama.manage import list_entity, stats
from jiama.rmq import Admin, Agent
from jiama.server import Rpc
from jiama.util import iter_dict, load_service, merge_dict


@click.group()
def jiama():
    pass


@jiama.command()
@click.argument('module_list', nargs=-1)
@click.option('--config', '-c', default='', help='The configuration file')
@click.option(
    '--broker', '-b', default='amqp://guest:guest@localhost', help='RabbitMQ broker uri'
)
def run(config, broker, module_list=[]):
    '''
    Run services. E.g.\n
    jiama run module:ServiceClass
    '''
    if config:
        with open(config, 'rb') as fp:
            config = tomli.load(fp)
    else:
        config = {'rpc': {'amqp_uri': broker}}
    log_config, rpc_config = merge_dict(DEFAULT_CONFIG, config).values()
    logger.configure(**log_config)

    rpc = Rpc(rpc_config)

    if '.' not in sys.path:
        sys.path.insert(0, '.')

    services = {}
    for m in module_list:
        services.update(load_service(m))
    asyncio.run(rpc.start(services))


@jiama.command()
@click.argument('entity', nargs=1, required=False)
@click.option(
    '--uri',
    '-u',
    default='http://guest:guest@localhost:15672/',
    help='RabbitMQ management API URI',
)
def list(uri, entity):
    '''
    List services or clients. E.g.\n
    jiama list [service | client] -u http://guest:guest@localhost:15672/
    '''
    logger.remove(handler_id=None)

    entity_list = asyncio.run(list_entity(uri, entity))
    for k, v in entity_list.items():
        click.secho(f'{k} list:'.capitalize(), fg='green', bold=True)
        for p in v:
            click.secho(f'    {p}', fg='green', bold=True)


@jiama.command()
@click.option(
    '--uri',
    '-u',
    default='http://guest:guest@localhost:15672/',
    help='RabbitMQ management API URI',
)
def status(uri):
    '''
    Show the RPC system status. E.g.\n
    jiama status -u http://guest:guest@localhost:15672/
    '''
    logger.remove(handler_id=None)

    result = asyncio.run(stats(uri))
    for i in iter_dict(result):
        click.secho(i, fg='blue', bold=True)



import asyncio
import json
import uuid

from loguru import logger

from jiama.constant import (
    DEFAULT_CONFIG, EXCHANGE_NAME, REQUEST_ROUTING_KEY, RESPONSE_BINDING_KEY,
    RESPONSE_QUEUE_NAME, RESPONSE_ROUTING_KEY
)
from jiama.rmq import Agent
from jiama.util import merge_dict, singleton


class RpcProxy:
    async def create(self, config: dict):
        rpc_config = merge_dict(DEFAULT_CONFIG, config, ignore=['log'])['rpc']
        self.client_id = rpc_config.get('client_id', None) or str(uuid.uuid4())
        self.timeout = rpc_config['timeout']
        self.agent = Agent().init(rpc_config)
        self.loop = asyncio.get_running_loop()
        self.futures = {}
        self.services = {}
        self.queue, self.consumer = await self.agent.subscribe(
            RESPONSE_QUEUE_NAME.format(client_id=self.client_id),
            RESPONSE_BINDING_KEY.format(client_id=self.client_id),
            self.on_response,
        )
        return self

    async def on_response(self, message):
        await message.ack()

        future = self.futures.pop(message.correlation_id, None)
        if future is None:
            return

        future.set_result(json.loads(message.body))

    def ensure_future(self, correlation_id):
        future = self.loop.create_future()
        self.futures[correlation_id] = future
        return future

    async def close(self):
        logger.info(f'Closing client: {self.client_id}')
        if self.consumer:
            await self.queue.cancel(self.consumer)
            del self.consumer
        await self.queue.unbind(EXCHANGE_NAME)
        for future in self.futures.values():
            if future.done():
                continue
            future.set_exception(asyncio.CancelledError)
        await self.queue.delete()
        await self.agent.close()

    def __getattr__(self, service_name):
        if service_name not in self.services:
            self.services[service_name] = Service(self, service_name)
        return self.services[service_name]

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()


class Service:
    def __init__(self, proxy, name):
        self.proxy = proxy
        self.name = name

    def __getattr__(self, method):
        return Method(self.proxy, self.name, method)


class Method:
    def __init__(self, proxy, service, name):
        self.proxy = proxy
        self.service = service
        self.name = name

    async def __call__(self, *args, **kwargs):
        payload = {'args': args, 'kwargs': kwargs}
        payload = json.dumps(payload).encode()
        routing_key = REQUEST_ROUTING_KEY.format(
            service_name=self.service, method_name=self.name
        )
        reply_to = RESPONSE_ROUTING_KEY.format(client_id=self.proxy.client_id)
        correlation_id = str(uuid.uuid4())

        logger.info('Requesting {} with {}', routing_key, payload)
        future = self.proxy.ensure_future(correlation_id)
        await self.proxy.agent.publish(payload, routing_key, reply_to, correlation_id)

        return await asyncio.wait_for(future, timeout=self.proxy.timeout)



from aio_pika import Message, connect_robust
from aio_pika.pool import Pool
from yujian.api import Client

from jiama.constant import CONTENT_ENCODING, CONTENT_TYPE, EXCHANGE_NAME, EXCHANGE_TYPE
from jiama.util import seperate_uri, singleton


class Agent:
    def init(self, config):
        self.uri = config['amqp_uri']
        self.max_connection = config['max_connection']
        self.max_channel = config['max_channel']
        self.prefetch_count = config['prefetch_count']
        self.delivery_mode = config['delivery_mode']
        self.durable = config['durable']
        self.timeout = config['timeout']
        self.expiration = config['expiration']
        self.conn_pool = Pool(self._connect, max_size=self.max_connection)
        self.channel_pool = Pool(self._channel, max_size=self.max_channel)
        return self

    async def _connect(self):
        return await connect_robust(self.uri, timeout=self.timeout)

    async def _channel(self):
        async with self.conn_pool.acquire() as conn:
            return await conn.channel()

    async def exchange(self):
        async with self.channel_pool.acquire() as channel:
            return await channel.declare_exchange(
                EXCHANGE_NAME,
                EXCHANGE_TYPE,
                durable=self.durable,
                auto_delete=False if self.durable else True,
            )

    async def queue(self, queue_name):
        async with self.channel_pool.acquire() as channel:
            await channel.set_qos(prefetch_count=self.prefetch_count)
            return await channel.declare_queue(
                queue_name,
                durable=self.durable,
                auto_delete=False if self.durable else True,
            )

    def message(self, body, reply_to, correlation_id, **kwargs):
        return Message(
            body,
            reply_to=reply_to,
            correlation_id=correlation_id,
            delivery_mode=self.delivery_mode,
            expiration=self.expiration,
            content_type=CONTENT_TYPE,
            content_encoding=CONTENT_ENCODING,
            **kwargs,
        )

    async def publish(
        self, payload, routing_key, reply_to=None, correlation_id=None, **kwargs
    ):
        '''
        如果是响应请求, routing_key 直接使用 reply_to
        另外，在实例化 Message 时 使用参数默认值和关键字参数 方便不需要传入值的参数
        '''
        exchange = await self.exchange()
        message = self.message(
            payload, reply_to=reply_to, correlation_id=correlation_id, **kwargs
        )
        await exchange.publish(message, routing_key=routing_key, timeout=self.timeout)

    async def subscribe(self, queue_name, binding_key, callback=None):
        exchange = await self.exchange()
        queue = await self.queue(queue_name)
        await queue.bind(exchange, binding_key)
        consumer = None
        if callback:
            consumer = await queue.consume(callback)
        return queue, consumer

    async def close(self):
        await self.channel_pool.close()
        await self.conn_pool.close()


class Admin:
    def init(self, api_uri):
        parts = seperate_uri(api_uri)
        self.uri = '{}{}'.format(parts.get('protocol'), parts.get('address'))
        self.user = parts.get('user')
        self.password = parts.get('password')
        return self

    async def overview(self):
        return await self.yujian.overview()

    async def list_queue(self, vhost: str = None, columns: list[str] = None, **kwargs):
        qs = await self.yujian.list_queue(vhost=vhost, columns=columns, **kwargs)
        return [q for q in qs if q['name'].startswith('Jiama')]

    async def destroy(self):
        await self.yujian.close()

    async def __aenter__(self):
        self.yujian = await Client().init(self.uri, self.user, self.password)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.destroy()



import sys


EXCHANGE_NAME = 'Jiama.rpc'
EXCHANGE_TYPE = 'topic'

REQUEST_QUEUE_NAME = 'Jiama.{service_name}'
REQUEST_BINDING_KEY = '{service_name}.*'
REQUEST_ROUTING_KEY = '{service_name}.{method_name}'

RESPONSE_QUEUE_NAME = 'Jiama.result.{client_id}'
RESPONSE_BINDING_KEY = '#.{client_id}'
RESPONSE_ROUTING_KEY = 'Jiama.result.{client_id}'

CONTENT_TYPE = 'application/json'
CONTENT_ENCODING = 'UTF8'

DEFAULT_CONFIG = {
    'rpc': {
        'amqp_uri': 'amqp://guest:guest@localhost',
        'max_channel': 10,
        'max_connection': 2,
        'prefetch_count': 5,
        'delivery_mode': 2,  # Marks a message as persistent (with a value of 2) or transient (any other value).
        'expiration': 180,  # expiration in seconds (or datetime or timedelta)
        'timeout': 5,  # connection timeout in seconds
        'durable': True,  # 队列和交换机都设置为durable的，这个设置将保证消息能够持久化
    },
    'log': {
        'handlers': [
            {
                'sink': sys.stdout,
                'level': 'DEBUG',
                'colorize': True,
                'format': '<green>{time:YYYY-MM-DD at HH:mm:ss}</green> {level}: <level>{message}</level>',
            },
        ]
    },
}



import asyncio
import json
import signal

from loguru import logger

from jiama.constant import EXCHANGE_NAME, REQUEST_BINDING_KEY, REQUEST_QUEUE_NAME
from jiama.rmq import Agent
from jiama.util import singleton


@singleton
class Rpc:
    def __init__(self, rpc_config):
        self.agent = Agent().init(rpc_config)
        self.listeners = []

    async def start(self, services):
        logger.info('Starting {}', list(services.keys()))

        loop = asyncio.get_running_loop()
        for s in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(self.shutdown(loop, s))
            )

        self.services = services
        try:
            await asyncio.wait(
                [asyncio.create_task(self.listen(s)) for s in self.services.keys()]
            )
        except asyncio.CancelledError:
            logger.info('Tasks were cancelled ')
        finally:
            logger.info("Successfully shutdown services")

    async def listen(self, service):
        queue, consumer = await self.agent.subscribe(
            REQUEST_QUEUE_NAME.format(service_name=service),
            REQUEST_BINDING_KEY.format(service_name=service),
            self.on_request,
        )
        self.listeners.append((queue, consumer))
        await asyncio.Future()

    async def on_request(self, message):
        await message.ack()

        service, method = message.routing_key.split('.')
        logger.info('Processing {}.{}', service, method)

        body = json.loads(message.body.decode())
        args = body['args']
        kwargs = body['kwargs']

        service_instance = self.services[service]()
        method_instance = getattr(service_instance, method)
        if asyncio.iscoroutinefunction(method_instance):
            result = await method_instance(*args, **kwargs)
        else:
            result = method_instance(*args, **kwargs)

        payload = json.dumps(result).encode()
        routing_key = message.reply_to

        logger.info('Sending result to {}', routing_key)
        await self.agent.publish(
            payload, routing_key=routing_key, correlation_id=message.correlation_id
        )

    async def shutdown(self, loop, signal):
        logger.info(f'Received exit signal {signal.name}')
        for queue, consumer in self.listeners:
            if consumer:
                await queue.cancel(consumer)
                del consumer
            await queue.unbind(EXCHANGE_NAME)
            await queue.delete()
        await self.agent.close()

        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        for t in tasks:
            t.cancel()
        logger.info(f'Cancelling {len(tasks)} outstanding tasks')
        await asyncio.gather(*tasks, return_exceptions=True)

        loop.stop()

    @staticmethod
    def decorator(method_func):
        if getattr(method_func, 'jiama_rpc_entrypoint', False):
            return
        setattr(method_func, 'jiama_rpc_entrypoint', True)
        return method_func


rpc = Rpc.decorator
